# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-17 11:52
Desc:
'''
import pandas as pd
from pynovice.geo_map import GeoMap
import webbrowser

if __name__ == '__main__':
    df_raw = pd.DataFrame(data= [[-0.4302744, 35.5331974],
                                   [-1.21167  , 36.8973031],
                                   [ 0.5217472, 35.255192 ],
                                   [-2.2326433, 38.0769783],
                                   [-1.0625246, 34.4734754]],
                           columns=['latitude', 'longitude']
                           )
    geo_map = GeoMap()
    geo_map.set_out_file(html_path='../data/geo.html')

    # 全流程运行
    # geo_map.run_geo_map(df_raw,web_open=True)

    # 详细的流程
    df_gps = geo_map.get_geo_reverse(df_raw) #获取geo
    geo_data = df_gps.groupby("admin1")['lon'].count() #要画图的数据 serios
    geo_map.plot_geo_summary(geo_data) #生成geo.html 文件
    webbrowser.open(url='file://' + geo_map.html_name, new=0, autoraise=True) #浏览器打开geo.html文件


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-07-28 15:07
Desc: 数据分布画图
'''


from .trajectory_playback import Trajectory as Trajectory_his
from .trajectory_playback_v2 import Trajectory as Trajectory


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-25 14:40
Desc:
derived from https://github.com/ShawnyXiao/TextClassification-Keras
'''


def start():
    print("import successful")



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-28 16:23
Desc:
    针对离散数据,我们需要将离散的枚举值替换成数值才可以用于计算.这些数值就是各个枚举值的权重.
    广义上讲,woe可以算是一种编码方式.
    IV	预测能力
    <0.03	无预测能力
    0.03~0.09	低
    0.1~0.29	中
    0.3~0.49	高
    >=0.5	极高
'''
import numpy as np
import pandas as pd


class WeightOfEvidence:
    def __init__(self,bad_target=1):
        self.bad_target = bad_target
        self.eps = 1e-4
        self.woe_card = None

    def fit(self,df_x,df_y):
        '''
        :param df_x: 特征字段
        :param df_y: 标签分类字段
        :return:
            返回woe详细信息
        '''
        self.cur_field = df_x.name or 'field'
        # |fields |bins |bad |good |bad_prob |good_prob |woe |iv
        bin = np.unique(df_x)
        good = np.array([np.logical_and(df_x == val, df_y != self.bad_target).sum() for val in bin])
        bad = np.array([np.logical_and(df_x == val, df_y == self.bad_target).sum() for val in bin])
        prob_good = good / float(np.sum(df_y != self.bad_target))
        prob_bad = bad / float(np.sum(df_y==self.bad_target))
        woe = np.log(np.maximum(prob_good,self.eps)/np.maximum(prob_bad,self.eps))
        iv = (prob_good-prob_bad)*woe
        df_info = pd.DataFrame(data=np.array([bin,good,bad,prob_good,prob_bad,woe,iv]).T,
                               columns=['bin','good','bad','good_prob','bad_prob','woe','iv'])
        df_info.insert(0, 'fields', self.cur_field)
        self.woe_card=df_info

    def transform(self,x):
        card = self.woe_card
        if type(x) in (pd.DataFrame,pd.Series):
            df_x = pd.DataFrame(x)
            df_x.columns = ['bin']
            out = pd.DataFrame(df_x).join(card.set_index('bin'), on='bin', how='left')
        else:
            out = card[card.bin == x].woe
        return out

    def fit_transform(self,df_x,df_y):
        self.fit(df_x,df_y)
        df_x = pd.DataFrame(df_x)
        df_x.columns=['bin']
        out = pd.DataFrame(df_x).join(self.woe_card.set_index('bin'),on='bin',how='left')
        return out


class WeightOfFeeEvidence:
    def __init__(self,bad_target=1):
        self.bad_target = bad_target
        self.eps = 1e-4
        self.woe_card = None

    def fit(self,df_x,df_fee_got,df_report_fee):
        '''
        :param df_x: 特征字段
        :param df_fee_got: 已赚保费
        :param df_report_fee: 赔付金额
        :return:
            返回woe详细信息
        '''
        self.cur_field = df_x.name or 'field'
        # |fields |bins |bad |good |bad_prob |good_prob |woe |iv
        bin = np.unique(df_x)
        good = np.array([((df_x == val)*df_fee_got).sum() for val in bin])
        bad = np.array([((df_x == val)*df_report_fee).sum() for val in bin])
        prob_good = good / float(np.sum(df_fee_got))
        prob_bad = bad / float(np.sum(df_report_fee))
        woe = np.log(np.maximum(prob_good,self.eps)/np.maximum(prob_bad,self.eps))
        iv = (prob_good-prob_bad)*woe
        df_info = pd.DataFrame(data=np.array([bin,good,bad,prob_good,prob_bad,woe,iv]).T,
                               columns=['bin','get_fee','report_fee','get_fee_prob','report_fee_prob','woe','iv'])
        df_info.insert(0, 'fields', self.cur_field)
        self.woe_card=df_info

    def transform(self,x):
        card = self.woe_card
        if type(x) in (pd.DataFrame,pd.Series):
            df_x = pd.DataFrame(x)
            df_x.columns = ['bin']
            out = pd.DataFrame(df_x).join(card.set_index('bin'), on='bin', how='left')
        else:
            out = card[card.bin == x].woe
        return out

    def fit_transform(self, df_x, df_fee_got, df_report_fee):
        self.fit(df_x, df_fee_got, df_report_fee)
        df_x = pd.DataFrame(df_x)
        df_x.columns = ['bin']
        out = pd.DataFrame(df_x).join(self.woe_card.set_index('bin'), on='bin', how='left')
        return out

if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 3, 3, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']

    df_field = pd.cut(df_field,bins=3)
    print(df_field)
    woe = WeightOfEvidence()
    aa = woe.fit_transform(df_field, df_label)
    bb = woe.transform(df_field[0])
    print(aa)
    print(bb)
    print(woe.woe_card)








# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2021-02-08 18:09
Desc:
map extends
base图：
    1.折线图
    2.柱状图
    3.树图(pass)
    4.地图(日历图)
    5.桑基图(数据流转)
    6.热力图
    7.关系图
组合图：
'''

import pandas as pd
import numpy as np
import re
import os
from pyecharts.charts import Bar, Line, Page, Grid, Tab
from pyecharts.globals import ThemeType
from pyecharts.components import Table
from pyecharts import options as opts
from pynovice.rule_engine import RuleParser
from pynovice.util import progress_bar

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))

def get_kde(x,data_array,bandwidth=0.1):
    '''
    核密度函数
    :param x:
    :param data_array:
    :param bandwidth:
    :return:
    '''
    def gauss(x):
        import math
        return (1/math.sqrt(2*math.pi))*math.exp(-0.5*(x**2))
    N=len(data_array)
    res=0
    if len(data_array)==0:
        return 0
    for i in range(len(data_array)):
        res += gauss((x-data_array[i])/bandwidth)
    res /= (N*bandwidth)
    return res

def clac_kde_fit_tmp(arr_x,bins=50):
    '''
    密度拟合分布(del)
    :param arr_x:
    :param bins:
    :return:
    '''
    bins = min(len(arr_x),bins)
    bandwidth=1.05*np.std(arr_x)*(len(arr_x)**(-1/5))
    x_data = np.linspace(min(arr_x),max(arr_x),bins)
    y_data = pd.cut(arr_x,bins).value_counts().sort_index().tolist()
    y_fit_data = [get_kde(x_data[i],arr_x,bandwidth) for i in range(x_data.shape[0])]
    return x_data,y_data,y_fit_data

def clac_kde_fit(arr_x,bins_list):
    '''
    密度拟合分布
    '''
    bandwidth=1.05*np.std(arr_x)*(len(arr_x)**(-1/5))
    y_data = [get_kde(bins_list[i],arr_x,bandwidth) for i in range(len(bins_list))]
    y_data = [x / sum(y_data) for x in y_data]
    return y_data


def plot_base_bar1(df: pd.DataFrame, title_name='', subtitle_name='', xaxis_name='', yaxis_name='',
                   stack_type=False,is_show_label=False, is_plot: bool = True):
    '''
        柱状图
    '''
    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)
    # plot
    _data = df.copy()
    _data.index = [str(i) for i in _data.index]
    c1 = Bar({'bg_color':'white'})
    c1.add_xaxis(_data.index.tolist())
    for icol in _data.columns:
        if stack_type:
            c1.add_yaxis(icol, _data[icol].values.tolist(),stack='stack1')
        else:
            c1.add_yaxis(icol, _data[icol].values.tolist())
    c1.set_global_opts(
        title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
        legend_opts=opts.LegendOpts(is_show=True),
        toolbox_opts=opts.ToolboxOpts(is_show=True),
        xaxis_opts=opts.AxisOpts(name=xaxis_name),
        yaxis_opts=opts.AxisOpts(name=yaxis_name),
        datazoom_opts=opts.DataZoomOpts(is_show=True)
    )
    c1.set_series_opts(label_opts=opts.LabelOpts(is_show=is_show_label))
    if is_plot:
        return c1.render_notebook()
    else:
        return c1


def plot_base_line1(df: pd.DataFrame, title_name='', subtitle_name='', xaxis_name='', yaxis_name='',
                    is_show_label=False, is_plot: bool = True):
    '''
        折线图
    '''
    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)
    # plot
    _data = df.copy()
    _data.index = [str(i) for i in _data.index]
    c1 = Line({'bg_color':'white'})
    c1.add_xaxis(_data.index.tolist())
    for icol in _data.columns:
        c1.add_yaxis(icol, _data[icol].values.tolist())
    c1.set_global_opts(
        title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
        legend_opts=opts.LegendOpts(is_show=True),
        toolbox_opts=opts.ToolboxOpts(is_show=True),
        xaxis_opts=opts.AxisOpts(name=xaxis_name),
        yaxis_opts=opts.AxisOpts(name=yaxis_name),
        datazoom_opts=opts.DataZoomOpts(is_show=True)
    )
    c1.set_series_opts(label_opts=opts.LabelOpts(is_show=is_show_label))
    if is_plot:
        return c1.render_notebook()
    else:
        return c1

def plot_base_grid1(chart1_top, chart_buttom,is_plot: bool = True):
    '''
        组合折线图(上下)
    '''
    grid = Grid(init_opts=opts.InitOpts(theme="white", bg_color='white', ))
    grid.add(chart1_top, grid_opts=opts.GridOpts(pos_bottom="50%"))
    grid.add(chart_buttom, grid_opts=opts.GridOpts(pos_top="65%"))

    if is_plot:
        return grid.render_notebook()
    else:
        return grid

def plot_base_grid2(df1: pd.DataFrame, df2: pd.DataFrame,chart_type='barline',stack_type='FalseFalse',
                    title_name='', subtitle_name='', xaxis_name='', yaxis_name1='', yaxis_name2='',
                    is_plot: bool = True):
    '''
    折线图(双坐标)
    '''
    assert chart_type in ('barbar','lineline','barline','linebar')
    assert stack_type in ('FalseFalse','TrueFalse','TrueTrue','FalseTrue')
    if isinstance(df1, pd.Series):
        df1 = pd.DataFrame(df1)
    if isinstance(df2, pd.Series):
        df2 = pd.DataFrame(df2)
    assert df1.index.to_list()==df2.index.to_list()
    _data1 = df1.copy()
    _data2 = df2.copy()
    # plot
    if chart_type in ('barline','barbar'):
        c1 = Bar({'bg_color':'white'})
    else:
        c1 = Line({'bg_color':'white'})
    c1.set_global_opts(
        title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
        legend_opts=opts.LegendOpts(is_show=True),
        toolbox_opts=opts.ToolboxOpts(is_show=True),
        xaxis_opts=opts.AxisOpts(name=xaxis_name),
        yaxis_opts=opts.AxisOpts(name=yaxis_name1),
        datazoom_opts=opts.DataZoomOpts(is_show=True)
    )
    c1.extend_axis(yaxis=opts.AxisOpts(name=yaxis_name2,position="right"))
    c1.add_xaxis([str(i) for i in _data1.index])
    for icol in _data1.columns:
        if stack_type in ('FalseFalse','FalseTrue'):
            c1.add_yaxis(icol, _data1[icol].values.tolist(),yaxis_index=0)
        else:
            c1.add_yaxis(icol, _data1[icol].values.tolist(),yaxis_index=0,stack='stack1')
    if chart_type in ('lineline','barline'):
        c2 = Line()
    else:
        c2 = Bar()
    c2.add_xaxis([str(i) for i in _data1.index])
    for icol in _data2.columns:
        if stack_type in ('FalseFalse','TrueFalse'):
            c2.add_yaxis(icol, _data2[icol].values.tolist(),yaxis_index=1)
        else:
            c2.add_yaxis(icol, _data2[icol].values.tolist(),yaxis_index=1,stack='stack1')
    c1.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
    c2.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
    c1.overlap(c2)
    if is_plot:
        return c1.render_notebook()
    else:
        return c1

# def plot_base_line2(df1: pd.DataFrame, df2: pd.DataFrame,
#                     title_name='', subtitle_name='', xaxis_name='', yaxis_name1='', yaxis_name2='',
#                     is_save: bool = False):
#     '''
#     折线图(双坐标)
#     '''
#     if isinstance(df1, pd.Series):
#         df1 = pd.DataFrame(df1)
#     if isinstance(df2, pd.Series):
#         df2 = pd.DataFrame(df2)
#     assert df1.index.to_list()==df2.index.to_list()
#     _data1 = df1.copy()
#     _data2 = df2.copy()
#     # plot
#     c1 = Line({'bg_color':'white'})
#     c1.set_global_opts(
#         title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
#         legend_opts=opts.LegendOpts(is_show=True),
#         toolbox_opts=opts.ToolboxOpts(is_show=True),
#         xaxis_opts=opts.AxisOpts(name=xaxis_name),
#         yaxis_opts=opts.AxisOpts(name=yaxis_name1),
#         datazoom_opts=opts.DataZoomOpts(is_show=True)
#     )
#     c1.extend_axis(yaxis=opts.AxisOpts(name=yaxis_name2,position="right"))
#     c1.add_xaxis([str(i) for i in _data1.index])
#     for icol in _data1.columns:
#         c1.add_yaxis(icol, _data1[icol].values.tolist(),yaxis_index=0)
#     for icol in _data2.columns:
#         c1.add_yaxis(icol, _data2[icol].values.tolist(),yaxis_index=1)
#     c1.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
#     if is_save:
#         return c1
#     else:
#         return c1.render_notebook()
#
# def plot_base_barline1(df1: pd.DataFrame, df2: pd.DataFrame,
#                     title_name='', subtitle_name='', xaxis_name='', yaxis_name1='', yaxis_name2='',
#                     is_save: bool = False):
#     '''
#     折线图(双坐标)
#     '''
#     if isinstance(df1, pd.Series):
#         df1 = pd.DataFrame(df1)
#     if isinstance(df2, pd.Series):
#         df2 = pd.DataFrame(df2)
#     assert df1.index.to_list()==df2.index.to_list()
#     _data1 = df1.copy()
#     _data2 = df2.copy()
#     # plot
#     c1 = Bar({'bg_color':'white'})
#     c1.set_global_opts(
#         title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
#         legend_opts=opts.LegendOpts(is_show=True),
#         toolbox_opts=opts.ToolboxOpts(is_show=True),
#         xaxis_opts=opts.AxisOpts(name=xaxis_name),
#         yaxis_opts=opts.AxisOpts(name=yaxis_name1),
#         datazoom_opts=opts.DataZoomOpts(is_show=True)
#     )
#     c1.extend_axis(yaxis=opts.AxisOpts(name=yaxis_name2,position="right"))
#     c1.add_xaxis([str(i) for i in _data1.index])
#     for icol in _data1.columns:
#         c1.add_yaxis(icol, _data1[icol].values.tolist(),yaxis_index=0)
#     c2=Line()
#     c2.add_xaxis([str(i) for i in _data1.index])
#     for icol in _data2.columns:
#         c2.add_yaxis(icol, _data2[icol].values.tolist(),yaxis_index=1)
#     c1.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
#     c2.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
#     c1.overlap(c2)
#     if is_save:
#         return c1
#     else:
#         return c1.render_notebook()

# def plot_base_grid1(df1: pd.DataFrame, df2: pd.DataFrame,
#                     title_name='', subtitle_name='', xaxis_name='', yaxis_name1='', yaxis_name2='',
#                     is_save: bool = False):
#     '''
#         组合折线图(上下)
#     '''
#     if isinstance(df1, pd.Series):
#         df1 = pd.DataFrame(df1)
#     if isinstance(df2, pd.Series):
#         df2 = pd.DataFrame(df2)
#     len_col = df2.columns.size
#
#     # plot1
#     _data = df1.copy()
#     c1 = Line({'bg_color':'white'})
#     c1.add_xaxis([str(i) for i in _data.index])
#     for icol in _data.columns.tolist():
#         c1.add_yaxis(str(icol), _data[icol].values.tolist())
#     c1.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
#     if (len_col > 5):
#         opt_leg = opts.LegendOpts(is_show=True, type_='scroll',
#                                   pos_left='25%', pos_right='25%')
#     else:
#         opt_leg = opts.LegendOpts(is_show=True)
#     c1.set_global_opts(
#         title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
#         legend_opts=opt_leg,
#         toolbox_opts=opts.ToolboxOpts(is_show=True),
#         xaxis_opts=opts.AxisOpts(name=xaxis_name),
#         yaxis_opts=opts.AxisOpts(name=yaxis_name1),
#         datazoom_opts=opts.DataZoomOpts(is_show=True),
#     )
#     # plot2
#     _data = df2.copy()
#     c2 = Line({'bg_color':'white'})
#     c2.add_xaxis([str(i) for i in _data.index])
#     for icol in _data.columns.tolist():
#         c2.add_yaxis(str(icol), _data[icol].values.tolist())
#     c2.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
#     c2.set_global_opts(
#         legend_opts=opts.LegendOpts(is_show=False),
#         toolbox_opts=opts.ToolboxOpts(is_show=True),
#         xaxis_opts=opts.AxisOpts(name=xaxis_name),
#         yaxis_opts=opts.AxisOpts(name=yaxis_name2),
#     )
#     grid = Grid(init_opts=opts.InitOpts(theme="white", bg_color='white', ))
#     grid.add(c1, grid_opts=opts.GridOpts(pos_bottom="50%"))
#     grid.add(c2, grid_opts=opts.GridOpts(pos_top="65%"))
#
#     if is_save:
#         return grid
#     else:
#         return grid.render_notebook()


if __name__ == '__main__':
    pass


import warnings

import numpy as np

try:
    import matplotlib.pyplot as pl
    import matplotlib
except ImportError:
    warnings.warn("matplotlib could not be loaded!")
    pass
from .utils._utils import safe_isinstance, format_value
from pynovice.score_card.src.card_explanations.utils import _colors as colors

labels = {
    'MAIN_EFFECT': "SHAP main effect value for\n%s",
    'INTERACTION_VALUE': "SHAP interaction value",
    'INTERACTION_EFFECT': "SHAP interaction value for\n%s and %s",
    'VALUE': "SHAP value (impact on model output)",
    'GLOBAL_VALUE': "mean(|SHAP value|) (average impact on model output magnitude)",
    'VALUE_FOR': "SHAP value for\n%s",
    'PLOT_FOR': "SHAP plot for %s",
    'FEATURE': "Feature %s",
    'FEATURE_VALUE': "Feature value",
    'FEATURE_VALUE_LOW': "Low",
    'FEATURE_VALUE_HIGH': "High",
    'JOINT_VALUE': "Joint SHAP value",
    'MODEL_OUTPUT': "Model output value"
}

# def waterfall(shap_values, max_display=10, show=True):
def waterfall(base_value, shap_values = None, features = None, feature_names = None, max_display=10, show=True):
    """
    
    Parameters
    ----------
    shap_values : Explanation
        A one-dimensional Explanation object that contains the feature values and SHAP values to plot.

    max_display : str
        The maximum number of features to plot.

    show : bool
        Whether matplotlib.pyplot.show() is called before returning. Setting this to False allows the plot
        to be customized further after it has been created.
    """

    base_values=base_value
    lower_bounds = getattr(shap_values, "lower_bounds", None)
    upper_bounds = getattr(shap_values, "upper_bounds", None)
    values = shap_values

    # make sure we only have a single output to explain
    if (type(base_values) == np.ndarray and len(base_values) > 0) or type(base_values) == list:
        raise Exception("waterfall_plot requires a scalar base_values of the model output as the first " \
                        "parameter, but you have passed an array as the first parameter! " \
                        "Try shap.waterfall_plot(explainer.base_values[0], values[0], X[0]) or " \
                        "for multi-output models try " \
                        "shap.waterfall_plot(explainer.base_values[0], values[0][0], X[0]).")

    # make sure we only have a single explanation to plot
    if len(values.shape) == 2:
        raise Exception("The waterfall_plot can currently only plot a single explanation but a matrix of explanations was passed!")
    
    # unwrap pandas series
    if safe_isinstance(features, "pandas.core.series.Series"):
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values

    # fallback feature names
    if feature_names is None:
        feature_names = np.array(["Feature %s" % str(i) for i in range(len(values))])
    
    # init variables we use for tracking the plot locations
    num_features = min(max_display, len(values))
    row_height = 0.5
    rng = range(num_features-1, -1, -1)
    order = np.argsort(-np.abs(values))
    pos_lefts = []
    pos_inds = []
    pos_widths = []
    pos_low = []
    pos_high = []
    neg_lefts = []
    neg_inds = []
    neg_widths = []
    neg_low = []
    neg_high = []
    loc = base_values + values.sum()
    yticklabels = ["" for i in range(num_features + 1)]
    
    # size the plot based on how many features we are plotting
    pl.gcf().set_size_inches(8, num_features * row_height + 1.5)
    pl.rcParams['font.sans-serif']=['SimHei']  # 用来正常显示中文标签

    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features == len(values):
        num_individual = num_features
    else:
        num_individual = num_features - 1

    # compute the locations of the individual features and plot the dashed connecting lines
    for i in range(num_individual):
        sval = values[order[i]]
        loc -= sval
        if sval >= 0:
            pos_inds.append(rng[i])
            pos_widths.append(sval)
            if lower_bounds is not None:
                pos_low.append(lower_bounds[order[i]])
                pos_high.append(upper_bounds[order[i]])
            pos_lefts.append(loc)
        else:
            neg_inds.append(rng[i])
            neg_widths.append(sval)
            if lower_bounds is not None:
                neg_low.append(lower_bounds[order[i]])
                neg_high.append(upper_bounds[order[i]])
            neg_lefts.append(loc)
        if num_individual != num_features or i + 4 < num_individual:
            pl.plot([loc, loc], [rng[i] -1 - 0.4, rng[i] + 0.4], color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
        if features is None:
            yticklabels[rng[i]] = feature_names[order[i]]
        else:
            yticklabels[rng[i]] = format_value(features[order[i]], "%0.03f") + " = " + feature_names[order[i]] 
    
    # add a last grouped feature to represent the impact of all the features we didn't show
    if num_features < len(values):
        yticklabels[0] = "%d other features" % (len(values) - num_features + 1)
        remaining_impact = base_values - loc
        if remaining_impact < 0:
            pos_inds.append(0)
            pos_widths.append(-remaining_impact)
            pos_lefts.append(loc + remaining_impact)
            c = colors.red_rgb
        else:
            neg_inds.append(0)
            neg_widths.append(-remaining_impact)
            neg_lefts.append(loc + remaining_impact)
            c = colors.blue_rgb

    points = pos_lefts + list(np.array(pos_lefts) + np.array(pos_widths)) + neg_lefts + list(np.array(neg_lefts) + np.array(neg_widths))
    dataw = np.max(points) - np.min(points)
    
    # draw invisible bars just for sizing the axes
    label_padding = np.array([0.1*dataw if w < 1 else 0 for w in pos_widths])
    pl.barh(pos_inds, np.array(pos_widths) + label_padding + 0.02 * dataw, left=np.array(pos_lefts) - 0.01*dataw, color=colors.red_rgb, alpha=0)
    label_padding = np.array([-0.1*dataw  if -w < 1 else 0 for w in neg_widths])
    pl.barh(neg_inds, np.array(neg_widths) + label_padding - 0.02 * dataw, left=np.array(neg_lefts) + 0.01*dataw, color=colors.blue_rgb, alpha=0)
    
    # define variable we need for plotting the arrows
    head_length = 0.08
    bar_width = 0.8
    xlen = pl.xlim()[1] - pl.xlim()[0]
    fig = pl.gcf()
    ax = pl.gca()
    xticks = ax.get_xticks()
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    bbox_to_xscale = xlen/width
    hl_scaled = bbox_to_xscale * head_length
    renderer = fig.canvas.get_renderer()
    
    # draw the positive arrows
    for i in range(len(pos_inds)):
        dist = pos_widths[i]
        arrow_obj = pl.arrow(
            pos_lefts[i], pos_inds[i], max(dist-hl_scaled, 0.000001), 0,
            head_length=min(dist, hl_scaled),
            color=colors.red_rgb, width=bar_width,
            head_width=bar_width
        )
        
        if pos_low is not None and i < len(pos_low):
            pl.errorbar(
                pos_lefts[i] + pos_widths[i], pos_inds[i], 
                xerr=np.array([[pos_widths[i] - pos_low[i]], [pos_high[i] - pos_widths[i]]]),
                ecolor=colors.light_red_rgb
            )
        txt_obj = pl.text(
            pos_lefts[i] + 0.5*dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)
        
        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width: 
            txt_obj.remove()
            
            txt_obj = pl.text(
                pos_lefts[i] + (5/72)*bbox_to_xscale + dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
                horizontalalignment='left', verticalalignment='center', color=colors.red_rgb,
                fontsize=12
            )
    
    # draw the negative arrows
    for i in range(len(neg_inds)):
        dist = neg_widths[i]
        
        arrow_obj = pl.arrow(
            neg_lefts[i], neg_inds[i], -max(-dist-hl_scaled, 0.000001), 0,
            head_length=min(-dist, hl_scaled),
            color=colors.blue_rgb, width=bar_width,
            head_width=bar_width
        )

        if neg_low is not None and i < len(neg_low):
            pl.errorbar(
                neg_lefts[i] + neg_widths[i], neg_inds[i], 
                xerr=np.array([[neg_widths[i] - neg_low[i]], [neg_high[i] - neg_widths[i]]]),
                ecolor=colors.light_blue_rgb
            )
        txt_obj = pl.text(
            neg_lefts[i] + 0.5*dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)
        
        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width: 
            txt_obj.remove()

            txt_obj = pl.text(
                neg_lefts[i] - (5/72)*bbox_to_xscale + dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
                horizontalalignment='right', verticalalignment='center', color=colors.blue_rgb,
                fontsize=12
            )

    # draw the y-ticks twice, once in gray and then again with just the feature names in black
    ytick_pos = list(range(num_features)) + list(np.arange(num_features)+1e-8) # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    pl.yticks(ytick_pos, yticklabels[:-1] + [l.split('=')[-1] for l in yticklabels[:-1]], fontsize=13)
    
    # put horizontal lines for each feature row
    for i in range(num_features):
        pl.axhline(i, color="#cccccc", lw=0.5, dashes=(1, 5), zorder=-1)
    
    # mark the prior expected value and the model prediction
    pl.axvline(base_values, 0, 1/num_features, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
    fx = base_values + values.sum()
    pl.axvline(fx, 0, 1, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
    
    # clean up the main axis
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    pl.gca().spines['left'].set_visible(False)
    ax.tick_params(labelsize=13)
    #pl.xlabel("\nModel output", fontsize=12)

    # draw the E[f(X)] tick mark
    xmin,xmax = ax.get_xlim()
    ax2=ax.twiny()
    ax2.set_xlim(xmin,xmax)
    ax2.set_xticks([base_values, base_values+1e-8]) # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax2.set_xticklabels(["\n$E[f(X)]$","\n$ = "+format_value(base_values, "%0.03f")+"$"], fontsize=12, ha="left")
    ax2.spines['right'].set_visible(False)
    ax2.spines['top'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # draw the f(x) tick mark
    ax3=ax2.twiny()
    ax3.set_xlim(xmin,xmax)
    ax3.set_xticks([base_values + values.sum(), base_values + values.sum() + 1e-8]) # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax3.set_xticklabels(["$f(x)$","$ = "+format_value(fx, "%0.03f")+"$"], fontsize=12, ha="left")
    tick_labels = ax3.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-10/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(12/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_color("#999999")
    ax3.spines['right'].set_visible(False)
    ax3.spines['top'].set_visible(False)
    ax3.spines['left'].set_visible(False)

    # adjust the position of the E[f(X)] = x.xx label
    tick_labels = ax2.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-20/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(22/72., -1/72., fig.dpi_scale_trans))
    
    tick_labels[1].set_color("#999999")

    # color the y tick labels that have the feature values as gray
    # (these fall behind the black ones with just the feature name)
    tick_labels = ax.yaxis.get_majorticklabels()
    for i in range(num_features):
        tick_labels[i].set_color("#999999")
    
    if show:
        pl.show()



def waterfall_legacy(base_value, shap_values=None, features=None, feature_names=None, max_display=10, show=True):
    """ Plots an explantion of a single prediction as a waterfall plot.

    The SHAP value of a feature represents the impact of the evidence provided by that feature on the model's
    output. The waterfall plot is designed to visually display how the SHAP values (evidence) of each feature
    move the model output from our prior expectation under the background data distribution, to the final model
    prediction given the evidence of all the features. Features are sorted by the magnitude of their SHAP values
    with the smallest magnitude features grouped together at the bottom of the plot when the number of features
    in the models exceeds the max_display parameter.
    
    Parameters
    ----------
    expected_value : float
        This is the reference value that the feature contributions start from. For SHAP values it should
        be the value of explainer.expected_value.

    shap_values : numpy.array
        One dimensional array of SHAP values.

    features : numpy.array
        One dimensional array of feature values. This provides the values of all the
        features, and should be the same shape as the shap_values argument.

    feature_names : list
        List of feature names (# features).

    max_display : str
        The maximum number of features to plot.

    show : bool
        Whether matplotlib.pyplot.show() is called before returning. Setting this to False allows the plot
        to be customized further after it has been created.
    """
    
    # support passing an explanation object
    upper_bounds = None
    lower_bounds = None
    expected_value=base_value
    if str(type(expected_value)).endswith("Explanation'>"):
        shap_exp = expected_value
        expected_value = shap_exp.expected_value
        shap_values = shap_exp.values
        features = shap_exp.data
        feature_names = shap_exp.feature_names
        lower_bounds = getattr(shap_exp, "lower_bounds", None)
        upper_bounds = getattr(shap_exp, "upper_bounds", None)

    # make sure we only have a single output to explain
    if (type(expected_value) == np.ndarray and len(expected_value) > 0) or type(expected_value) == list:
        raise Exception("waterfall_plot requires a scalar expected_value of the model output as the first " \
                        "parameter, but you have passed an array as the first parameter! " \
                        "Try shap.waterfall_plot(explainer.expected_value[0], shap_values[0], X[0]) or " \
                        "for multi-output models try " \
                        "shap.waterfall_plot(explainer.expected_value[0], shap_values[0][0], X[0]).")

    # make sure we only have a single explanation to plot
    if len(shap_values.shape) == 2:
        raise Exception("The waterfall_plot can currently only plot a single explanation but a matrix of explanations was passed!")
    
    # unwrap pandas series
    if safe_isinstance(features, "pandas.core.series.Series"):
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values

    # fallback feature names
    if feature_names is None:
        feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(shap_values))])
    
    # init variables we use for tracking the plot locations
    num_features = min(max_display, len(shap_values))
    row_height = 0.5
    rng = range(num_features - 1, -1, -1)
    order = np.argsort(-np.abs(shap_values))
    pos_lefts = []
    pos_inds = []
    pos_widths = []
    pos_low = []
    pos_high = []
    neg_lefts = []
    neg_inds = []
    neg_widths = []
    neg_low = []
    neg_high = []
    loc = expected_value + shap_values.sum()
    yticklabels = ["" for i in range(num_features + 1)]
    
    # size the plot based on how many features we are plotting
    pl.gcf().set_size_inches(8, num_features * row_height + 1.5)
    # pl.rcParams['font.sans-serif'] = ['SimSun']

    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features == len(shap_values):
        num_individual = num_features
    else:
        num_individual = num_features - 1

    # compute the locations of the individual features and plot the dashed connecting lines
    for i in range(num_individual):
        sval = shap_values[order[i]]
        loc -= sval
        if sval >= 0:
            pos_inds.append(rng[i])
            pos_widths.append(sval)
            if lower_bounds is not None:
                pos_low.append(lower_bounds[order[i]])
                pos_high.append(upper_bounds[order[i]])
            pos_lefts.append(loc)
        else:
            neg_inds.append(rng[i])
            neg_widths.append(sval)
            if lower_bounds is not None:
                neg_low.append(lower_bounds[order[i]])
                neg_high.append(upper_bounds[order[i]])
            neg_lefts.append(loc)
        if num_individual != num_features or i + 4 < num_individual:
            pl.plot([loc, loc], [rng[i] -1 - 0.4, rng[i] + 0.4], color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
        if features is None:
            yticklabels[rng[i]] = feature_names[order[i]]
        else:
            yticklabels[rng[i]] = format_value(features[order[i]], "%0.03f") + " = " + feature_names[order[i]] 
    
    # add a last grouped feature to represent the impact of all the features we didn't show
    if num_features < len(shap_values):
        yticklabels[0] = "%d other features" % (len(shap_values) - num_features + 1)
        remaining_impact = expected_value - loc
        if remaining_impact < 0:
            pos_inds.append(0)
            pos_widths.append(-remaining_impact)
            pos_lefts.append(loc + remaining_impact)
            c = colors.red_rgb
        else:
            neg_inds.append(0)
            neg_widths.append(-remaining_impact)
            neg_lefts.append(loc + remaining_impact)
            c = colors.blue_rgb

    points = pos_lefts + list(np.array(pos_lefts) + np.array(pos_widths)) + neg_lefts + list(np.array(neg_lefts) + np.array(neg_widths))
    dataw = np.max(points) - np.min(points)
    
    # draw invisible bars just for sizing the axes
    label_padding = np.array([0.1*dataw if w < 1 else 0 for w in pos_widths])
    pl.barh(pos_inds, np.array(pos_widths) + label_padding + 0.02 * dataw, left=np.array(pos_lefts) - 0.01*dataw, color=colors.red_rgb, alpha=0)
    label_padding = np.array([-0.1*dataw  if -w < 1 else 0 for w in neg_widths])
    pl.barh(neg_inds, np.array(neg_widths) + label_padding - 0.02 * dataw, left=np.array(neg_lefts) + 0.01*dataw, color=colors.blue_rgb, alpha=0)
    
    # define variable we need for plotting the arrows
    head_length = 0.08
    bar_width = 0.8
    xlen = pl.xlim()[1] - pl.xlim()[0]
    fig = pl.gcf()
    ax = pl.gca()
    xticks = ax.get_xticks()
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    bbox_to_xscale = xlen/width
    hl_scaled = bbox_to_xscale * head_length
    renderer = fig.canvas.get_renderer()
    
    # draw the positive arrows
    for i in range(len(pos_inds)):
        dist = pos_widths[i]
        arrow_obj = pl.arrow(
            pos_lefts[i], pos_inds[i], max(dist-hl_scaled, 0.000001), 0,
            head_length=min(dist, hl_scaled),
            color=colors.red_rgb, width=bar_width,
            head_width=bar_width
        )
        
        if pos_low is not None and i < len(pos_low):
            pl.errorbar(
                pos_lefts[i] + pos_widths[i], pos_inds[i], 
                xerr=np.array([[pos_widths[i] - pos_low[i]], [pos_high[i] - pos_widths[i]]]),
                ecolor=colors.light_red_rgb
            )

        txt_obj = pl.text(
            pos_lefts[i] + 0.5*dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)
        
        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width: 
            txt_obj.remove()
            
            txt_obj = pl.text(
                pos_lefts[i] + (5/72)*bbox_to_xscale + dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
                horizontalalignment='left', verticalalignment='center', color=colors.red_rgb,
                fontsize=12
            )
    
    # draw the negative arrows
    for i in range(len(neg_inds)):
        dist = neg_widths[i]
        
        arrow_obj = pl.arrow(
            neg_lefts[i], neg_inds[i], -max(-dist-hl_scaled, 0.000001), 0,
            head_length=min(-dist, hl_scaled),
            color=colors.blue_rgb, width=bar_width,
            head_width=bar_width
        )

        if neg_low is not None and i < len(neg_low):
            pl.errorbar(
                neg_lefts[i] + neg_widths[i], neg_inds[i], 
                xerr=np.array([[neg_widths[i] - neg_low[i]], [neg_high[i] - neg_widths[i]]]),
                ecolor=colors.light_blue_rgb
            )
        
        txt_obj = pl.text(
            neg_lefts[i] + 0.5*dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)
        
        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width: 
            txt_obj.remove()
            
            txt_obj = pl.text(
                neg_lefts[i] - (5/72)*bbox_to_xscale + dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
                horizontalalignment='right', verticalalignment='center', color=colors.blue_rgb,
                fontsize=12
            )

    # draw the y-ticks twice, once in gray and then again with just the feature names in black
    pl.yticks(list(range(num_features))*2, yticklabels[:-1] + [l.split('=')[-1] for l in yticklabels[:-1]], fontsize=13)
    
    # put horizontal lines for each feature row
    for i in range(num_features):
        pl.axhline(i, color="#cccccc", lw=0.5, dashes=(1, 5), zorder=-1)
    
    # mark the prior expected value and the model prediction
    pl.axvline(expected_value, 0, 1/num_features, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
    fx = expected_value + shap_values.sum()
    pl.axvline(fx, 0, 1, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
    
    # clean up the main axis
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    pl.gca().spines['left'].set_visible(False)
    ax.tick_params(labelsize=13)
    #pl.xlabel("\nModel output", fontsize=12)

    # draw the E[f(X)] tick mark
    xmin,xmax = ax.get_xlim()
    ax2=ax.twiny()
    ax2.set_xlim(xmin,xmax)
    ax2.set_xticks([expected_value, expected_value])
    ax2.set_xticklabels(["\n$E[f(X)]$","\n$ = "+format_value(expected_value, "%0.03f")+"$"], fontsize=12, ha="left")
    ax2.spines['right'].set_visible(False)
    ax2.spines['top'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # draw the f(x) tick mark
    ax3=ax2.twiny()
    ax3.set_xlim(xmin,xmax)
    ax3.set_xticks([expected_value + shap_values.sum()] * 2)
    ax3.set_xticklabels(["$f(x)$","$ = "+format_value(fx, "%0.03f")+"$"], fontsize=12, ha="left")
    tick_labels = ax3.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-10/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(12/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_color("#999999")
    ax3.spines['right'].set_visible(False)
    ax3.spines['top'].set_visible(False)
    ax3.spines['left'].set_visible(False)

    # adjust the position of the E[f(X)] = x.xx label
    tick_labels = ax2.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-20/72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(22/72., -1/72., fig.dpi_scale_trans))
    tick_labels[1].set_color("#999999")

    # color the y tick labels that have the feature values as gray
    # (these fall behind the black ones with just the feature name)
    tick_labels = ax.yaxis.get_majorticklabels()
    for i in range(num_features):
        tick_labels[i].set_color("#999999")
    
    if show:
        pl.show()


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-25 14:40
Desc:
'''

from .util import progress_bar
from .util import get_retrieve_name
from .util import func_check_vin
from .util import get_fee_month


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-27 15:17
Desc:
'''

import sys
import math
import inspect
import signal

# 1.进度条展示 progress_bar(1, 100)
def progress_bar(portion, total):
    """
    total 总数据大小，portion 已经传送的数据大小
    :param portion: 已经接收的数据量
    :param total: 总数据量
    :return: 接收数据完成，返回True
    """
    part = total / 50  # 1%数据的大小
    count = math.ceil(portion / part)
    sys.stdout.write('\r')
    sys.stdout.write(('[%-50s]%.2f%%' % (('>' * count), portion / total * 100)))
    sys.stdout.flush()

    if portion >= total:
        sys.stdout.write('\n')
        return True
    return False

# 2.获取变量名
def get_retrieve_name(var):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    list_name = [var_name for var_name, var_val in callers_local_vars if var_val is var]
    return list_name

# 3.函数超时退出
def set_timeout(seconds, callback_func):
    '''
    暂时不可用
    :param seconds: 超时时间秒
    :param callback_func:超时后执行的函数
    :return:
    '''
    def wrap(func):
        def handle(signum, frame):  # 收到信号 SIGALRM 后的回调函数，第一个参数是信号的数字，第二个参数是the interrupted stack frame.
            raise RuntimeError

        def to_do(*args, **kwargs):
            try:
                signal.signal(signal.SIGALRM, handle)  # 设置信号和回调函数
                signal.alarm(seconds)  # 设置 num 秒的闹钟
                # print('start alarm signal.')
                r = func(*args, **kwargs)
                # print('close alarm signal.')
                signal.alarm(0)  # 关闭闹钟
                return r
            except RuntimeError as e:
                if callback_func:
                    callback_func()

        return to_do
    return wrap


## vin 校准（检验车辆识别码是否正确）
def func_check_vin(vin):
    # https://www.angelfire.com/ca/TORONTO/VIN/VDS.html#equip
    if len(vin) != 17:
        return False
    if len(set(vin[:8] + vin[9:])) < 2:
        return False
    vin_ord = [ord(i) for i in vin]

    # 1.校验位(第9位)
    if (vin_ord[8] >= 48 and vin_ord[8] <= 57):  # 0到9
        verify = vin_ord[8] - 48
    elif (vin_ord[8] == 88):  # X
        verify = 10;
    else:
        return False

    # 位置权重
    weight = [8, 7, 6, 5, 4, 3, 2, 10, 0, 9, 8, 7, 6, 5, 4, 3, 2]

    # 字符对应值
    def get_symbol_trans(_ord_str):
        if (_ord_str >= 48 and _ord_str <= 57):  # 0-9
            symbol = _ord_str - 48
        elif (_ord_str >= 65 and _ord_str <= 72):  # A-H
            symbol = _ord_str - 64
        elif (_ord_str >= 74 and _ord_str <= 82 and _ord_str != 79 and _ord_str != 81):  # J-R不含O,Q
            symbol = _ord_str - 73
        elif (_ord_str >= 83 and _ord_str <= 90):  # S-Z
            symbol = _ord_str - 81
        else:
            return -999
        return symbol

    total_sum = 0
    for i, iv in enumerate(vin_ord):
        _symbol = get_symbol_trans(iv)
        if _symbol < 0:
            return False
        total_sum += _symbol * weight[i]

    return verify == total_sum % 11

## 房贷利率
def get_fee_month(loan_amount, year_rates, _type=1, pay_month=360):
    '''
    :param loan_amount: 贷款金额
    :param year_rates: 年化利率=月利率*12
    :param _type: {1:'等额本息',2:'等额本金',3:'先息后本'}
    :param pay_month: 1期为1月，30年即为360期
    :return: 月还款明细
    '''
    '''
    等额本息计算公式
    每月还款额=贷款本金×[月利率×（1+月利率）^还款月数]÷[（1+月利率）^还款月数-1]
    总支付利息：总利息=还款月数×每月月供额-贷款本金
    每月应还利息=贷款本金×月利率×〔(1+月利率)^还款月数-(1+月利率)^(还款月序号-1)〕÷〔(1+月利率)^还款月数-1〕
    每月应还本金=贷款本金×月利率×(1+月利率)^(还款月序号-1)÷〔(1+月利率)^还款月数-1〕
    总利息=还款月数×每月月供额-贷款本金
    
    等额本金计算公式
    每月月供额=(贷款本金÷还款月数)+(贷款本金-已归还本金累计额)×月利率
    每月应还本金=贷款本金÷还款月数
    每月应还利息=剩余本金×月利率=(贷款本金-已归还本金累计额)×月利率。
    每月月供递减额=每月应还本金×月利率=贷款本金÷还款月数×月利率
    总利息=还款月数×(总贷款额×月利率-月利率×(总贷款额÷还款月数)*(还款月数-1)÷2+总贷款额÷还款月数)
    '''
    month_rates = year_rates / 12
    _col = ['期数', '还款总额', '还款本金', '还款利息']
    if 1 == _type:  # 等额本息
        data = []
        for month_i in range(pay_month):
            interest = loan_amount * month_rates * ((1 + month_rates) ** pay_month - (1 + month_rates) ** month_i) / (
            (1 + month_rates) ** pay_month - 1)
            principal = loan_amount * month_rates * (1 + month_rates) ** month_i / ((1 + month_rates) ** pay_month - 1)
            _data = [month_i + 1, round(interest + principal, 2), round(principal, 2), round(interest, 2)]
            data.append(_data)
        df = pd.DataFrame(data, columns=_col)
    elif 2 == _type:  # 等额本金
        data = []
        payed_fee = 0
        principal = loan_amount / pay_month
        for month_i in range(pay_month):
            payed_fee = payed_fee +  principal
            interest = (loan_amount - payed_fee) * month_rates
            _data = [month_i + 1, round(interest + principal, 2), round(principal, 2), round(interest, 2)]
            data.append(_data)
        df = pd.DataFrame(data, columns=_col)
    else:  # 先息后本
        data = []
        interest = loan_amount * month_rates
        for month_i in range(pay_month - 1):
            _data = [month_i + 1, round(interest, 2), 0, round(interest, 2)]
            data.append(_data)
        data.append([month_i + 2, round(interest + loan_amount, 2), round(loan_amount, 2), round(interest, 2)])
        df = pd.DataFrame(data, columns=_col)
    return df



## 手机号正则
mobile_rg=r'^1[3|4|5|6|7|8|9][0-9]{9}$'
ydmobile_rg=r'^1(3[4-9]|4[7]|5[012789]|7[28]|8[23478]|9[578])\d{8}$'
ltmobile_rg=r'^1(3[0-2]|4[56]|5[56]|6[6]|7[0156]|8[56]|9[6])\d{8}$'
dxmobile_rg=r'^1(3[3]|4[19]|5[3]|7[3479]|8[019]|9[0139])\d{8}$'
def re_mobile(mobile):
    if re.match(ydmobile_rg,str(mobile)):
        return "移动"
    elif re.match(ltmobile_rg,str(mobile)):
        return "联通"
    elif re.match(mobile_rg,str(mobile)):
        return "电信"
    elif re.match(mobile_rg,str(mobile)):
        return "未知"
    else:
        return "非法手机号"

if __name__ == '__main__':
    ## 1.进度条
    portion = 0
    total = 254820000
    while True:
        portion += 1024
        sum = progress_bar(portion, total)
        if sum:
           break
    print("ok")
    # or
    progress_bar(portion=32, total=100)


    ## 2.变量名
    name = "bob"
    age = "23"
    bb = (name,age)
    for i in bb:
        var = get_retrieve_name(i)
        print(var)

    ## 3.函数超时
    def aa():
        print("超时")
    @set_timeout(2, callback_func=aa)
    def tt1():
        import time
        time.sleep(3)
    tt1()



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-16 16:51
Desc:
'''
from pynovice.bert_extracter import DownloadBertModel,ExtractBertFeatures

if __name__ == '__main__':
    # aa=DownloadBertModel()
    # aa.dump()
    bb=ExtractBertFeatures()
    print(bb.predict('你好啊'))


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-28 17:26
Desc:
    于sklearn决策树的最优分箱

'''

from sklearn.tree import DecisionTreeClassifier
import numpy as np

def tree_blocks(df_field,df_label,box_num=5,criterion='entropy'):
    '''
        利用决策树获得最优分箱的边界值列表
    '''
    if box_num<2:
        return [-np.inf,np.inf]
    internal_nodes = box_num - 1

    boundary = []
    x = df_field.fillna(-999).values  # 填充缺失值
    y = df_label.values
    clf = DecisionTreeClassifier(criterion=criterion,  # “信息熵”最小化准则划分,gini,选择信息熵更合理。Gini 指数更偏向于连续属性，熵更偏向于离散属性。
                                 max_leaf_nodes=internal_nodes+1,  # 最大叶子节点数 = 非叶子结点+1
                                 min_samples_leaf=0.05)  # 叶子节点样本数量最小占比

    clf.fit(x.reshape(-1, 1), y)  # 训练决策树

    n_nodes = clf.tree_.node_count
    children_left = clf.tree_.children_left
    children_right = clf.tree_.children_right
    threshold = clf.tree_.threshold

    for i in range(n_nodes):
        if children_left[i] != children_right[i]:  # 获得决策树节点上的划分边界值(除叶子结点)
            boundary.append(threshold[i])
    boundary.sort()

    blocks = np.concatenate([[-np.inf], boundary, [np.inf]])

    return blocks.tolist()

# CRITERIA_CLF = {"gini": _criterion.Gini, "entropy": _criterion.Entropy}
# CRITERIA_REG = {"mse": _criterion.MSE, "friedman_mse": _criterion.FriedmanMSE,
#                 "mae": _criterion.MAE}

if __name__ == '__main__':
    import pandas as pd
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']
    aa = tree_blocks(df_field, df_label,box_num=2)
    print(aa)


# https://www.jianshu.com/p/0436ea8336db
# http://www.cocoachina.com/articles/87951
#
# eg:
# 魔法脚本路径pwd:
# /home/jumproot/.ipython/profile_default/startup/hive_magic.py

#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019/11/22 下午3:56
# @Title   : Jupyter notebook自定义hive读取数据到df的magic方法
# @Link    : https://ipython.readthedocs.io/en/stable/config/custommagics.html


from IPython.core import magic_arguments
from IPython.core.magic import line_magic, cell_magic, Magics, magics_class
from pyhive import hive
import pandas as pd
import subprocess


class HiveEngine(object):
    def __init__(self,hostname):
        _config={
                'host': '172.18.33.{}'.format(hostname),
                'port': 10000,
        	}
        self.cur = hive.connect(**_config).cursor()


    def execute(self, sql, size=None):
        for isql in sql.split(';'):
            self.cur.execute(isql)
        if size:
            rows = self.cur.fetchmany(size=size)
        else:
            rows = self.cur.fetchall()
        columns = [x[0] for x in self.cur.description]
        return columns, rows


    def close(self):
        self.cur.close()


def _exec_cmd(cmd):
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if stdout:
        print(stdout)
    if stderr:
        print(stderr)


@magics_class
class HiveMagic(Magics):

    def get_df(self, cell, size=None):
        hv = HiveEngine()
        columns, records = hv.execute(cell, size=size)
        hv.close()
        df = pd.DataFrame(records, columns=columns)
        return df


    def check_csv(self, df_cols, table):
        df_check = self.get_df("select * from {} limit 0".format(table))
        return df_cols == df_check.columns.to_list()


    @cell_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument("variable", help="the df variable name")
    @magic_arguments.argument("-n", default=115,help="select the host server")
    @magic_arguments.argument("-h", "--head", action="store_true", help="show 5 rows")
    @magic_arguments.argument("-f", "--file", action="store_true", help="absolute path of the file")
    def hive(self, line="", cell=None):
        args = magic_arguments.parse_argstring(self.hive, line)
        size = 5 if args.head else None
        file = args.file
        hostname=args.n
        if file:
            with open(cell.replace('\n','')) as f:
                cell = f.read().strip()
                if cell.endswith(";"):
                    cell = cell[:-1]
        hv = HiveEngine(hostname)
        columns, records = hv.execute(cell, size=size)
        hv.close()
        df = pd.DataFrame(records, columns=columns)
        self.shell.user_ns[args.variable] = df


    @cell_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument("variable", help="the df variable name")
    @magic_arguments.argument("-o", "--overwrite", action="store_true", help="overwrite or not")
    def hive_write(self, line, cell):
        args = magic_arguments.parse_argstring(self.hive_write, line)
        variable = args.variable
        overwrite = args.overwrite
        df = self.shell.user_ns[variable]

        cols = df.columns.to_list()
        if overwrite:
            drop_cmd = """hive -e 'drop table {table}'""".format(table=cell)
            create_cmd = """hive -e 'create table {table} ({fields}) row format delimited fields terminated by ",";'""" \
                .format(table=cell, fields=",".join([x + " string" for x in cols]))
            _exec_cmd(drop_cmd); _exec_cmd(create_cmd)

        # if self.check_csv(cols, cell):
        filename = "/home/jumproot/.jupyter/jupyter_ouput_data/{}.csv".format(cell)
        df.to_csv(filename, header=False, index=False, encoding="utf-8")
        write_cmd = """hive -e 'LOAD DATA LOCAL INPATH "{filename}" INTO TABLE {table}';""" \
            .format(filename=filename, table=cell)
        _exec_cmd(write_cmd)
        # else:
        #     print("Columns do not match.")


ip = get_ipython()
ip.register_magics(HiveMagic)




####
# import pymysql
# import pandas as pd
#
#
# class MySQLEngine(object):
#     def __init__(self, _db='ics_testing'):
#         self.conn = pymysql.connect(
#             host="rm-2ze79p6j8rq9m5nn4.mysql.rds.aliyuncs.com",
#             user='ics_testing_rw',
#             password='ics_testing@2021',
#             db=_db,
#             charset='utf8mb4'
#         )
#
#     def execute(self, sql, size=None):
#         with self.conn.cursor(cursor=pymysql.cursors.DictCursor) as cursor:
#             res = cursor.execute(sql)
#             if size:
#                 rows = cursor.fetchmany(size=size)
#             else:
#                 rows = cursor.fetchall()
#             return rows, res
#
#     def close(self):
#         self.conn.close()
#
#
# def run_sql(_sql, size=None):
#     mysql = MySQLEngine()
#     rows, records = mysql.execute(_sql)
#     mysql.close()
#     df = pd.DataFrame(rows)
#     return df